
import logging
from enum import Enum
from typing import List, Protocol

from drt_sdk_core.interfaces import IAddress
from drt_sdk_core.serializer import arg_to_string
from drt_sdk_core.transaction import Transaction
from drt_sdk_core.transaction_factories.transaction_builder import \
    TransactionBuilder

logger = logging.getLogger(__name__)


class IConfig(Protocol):
    chain_id: str
    min_gas_limit: int
    gas_limit_per_byte: int
    gas_limit_issue: int
    gas_limit_toggle_burn_role_globally: int
    gas_limit_dcdt_local_mint: int
    gas_limit_dcdt_local_burn: int
    gas_limit_set_special_role: int
    gas_limit_pausing: int
    gas_limit_freezing: int
    gas_limit_wiping: int
    gas_limit_dcdt_nft_create: int
    gas_limit_dcdt_nft_update_attributes: int
    gas_limit_dcdt_nft_add_quantity: int
    gas_limit_dcdt_nft_burn: int
    gas_limit_store_per_byte: int
    issue_cost: int
    dcdt_contract_address: IAddress


class RegisterAndSetAllRolesTokenType(Enum):
    NFT = "NFT"
    SFT = "SFT"
    META = "META"
    FNG = "FNG"


class TokenManagementTransactionsFactory:
    def __init__(self, config: IConfig):
        self._config = config
        self._true_as_hex = arg_to_string("true")
        self._false_as_hex = arg_to_string("false")

    def create_transaction_for_issuing_fungible(
        self,
        sender: IAddress,
        token_name: str,
        token_ticker: str,
        initial_supply: int,
        num_decimals: int,
        can_freeze: bool,
        can_wipe: bool,
        can_pause: bool,
        can_change_owner: bool,
        can_upgrade: bool,
        can_add_special_roles: bool
    ) -> Transaction:
        self._notify_about_unsetting_burn_role_globally()

        parts: List[str] = [
            "issue",
            arg_to_string(token_name),
            arg_to_string(token_ticker),
            arg_to_string(initial_supply),
            arg_to_string(num_decimals),
            *([arg_to_string("canFreeze"), self._true_as_hex if can_freeze else self._false_as_hex]),
            *([arg_to_string("canWipe"), self._true_as_hex if can_wipe else self._false_as_hex]),
            *([arg_to_string("canPause"), self._true_as_hex if can_pause else self._false_as_hex]),
            *([arg_to_string("canChangeOwner"), self._true_as_hex if can_change_owner else self._false_as_hex]),
            *([arg_to_string("canUpgrade"), self._true_as_hex if can_upgrade else self._false_as_hex]),
            *([arg_to_string("canAddSpecialRoles"), self._true_as_hex if can_add_special_roles else self._false_as_hex])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=self._config.issue_cost,
            gas_limit=self._config.gas_limit_issue,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def _notify_about_unsetting_burn_role_globally(self) -> None:
        logger.info("""
==========
IMPORTANT!
==========
You are about to issue (register) a new token. This will set the role "DCDTRoleBurnForAll" (globally).
Once the token is registered, you can unset this role by calling "unsetBurnRoleGlobally" (in a separate transaction).""")

    def create_transaction_for_issuing_semi_fungible(
        self,
        sender: IAddress,
        token_name: str,
        token_ticker: str,
        can_freeze: bool,
        can_wipe: bool,
        can_pause: bool,
        can_transfer_nft_create_role: bool,
        can_change_owner: bool,
        can_upgrade: bool,
        can_add_special_roles: bool
    ) -> Transaction:
        self._notify_about_unsetting_burn_role_globally()

        parts: List[str] = [
            "issueSemiFungible",
            arg_to_string(token_name),
            arg_to_string(token_ticker),
            *([arg_to_string("canFreeze"), self._true_as_hex if can_freeze else self._false_as_hex]),
            *([arg_to_string("canWipe"), self._true_as_hex if can_wipe else self._false_as_hex]),
            *([arg_to_string("canPause"), self._true_as_hex if can_pause else self._false_as_hex]),
            *([arg_to_string("canTransferNFTCreateRole"), self._true_as_hex if can_transfer_nft_create_role else self._false_as_hex]),
            *([arg_to_string("canChangeOwner"), self._true_as_hex if can_change_owner else self._false_as_hex]),
            *([arg_to_string("canUpgrade"), self._true_as_hex if can_upgrade else self._false_as_hex]),
            *([arg_to_string("canAddSpecialRoles"), self._true_as_hex if can_add_special_roles else self._false_as_hex])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=self._config.issue_cost,
            gas_limit=self._config.gas_limit_issue,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_issuing_non_fungible(
        self,
        sender: IAddress,
        token_name: str,
        token_ticker: str,
        can_freeze: bool,
        can_wipe: bool,
        can_pause: bool,
        can_transfer_nft_create_role: bool,
        can_change_owner: bool,
        can_upgrade: bool,
        can_add_special_roles: bool
    ) -> Transaction:
        self._notify_about_unsetting_burn_role_globally()

        parts: List[str] = [
            "issueNonFungible",
            arg_to_string(token_name),
            arg_to_string(token_ticker),
            *([arg_to_string("canFreeze"), self._true_as_hex if can_freeze else self._false_as_hex]),
            *([arg_to_string("canWipe"), self._true_as_hex if can_wipe else self._false_as_hex]),
            *([arg_to_string("canPause"), self._true_as_hex if can_pause else self._false_as_hex]),
            *([arg_to_string("canTransferNFTCreateRole"), self._true_as_hex if can_transfer_nft_create_role else self._false_as_hex]),
            *([arg_to_string("canChangeOwner"), self._true_as_hex if can_change_owner else self._false_as_hex]),
            *([arg_to_string("canUpgrade"), self._true_as_hex if can_upgrade else self._false_as_hex]),
            *([arg_to_string("canAddSpecialRoles"), self._true_as_hex if can_add_special_roles else self._false_as_hex])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=self._config.issue_cost,
            gas_limit=self._config.gas_limit_issue,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_registering_meta_dcdt(
        self,
        sender: IAddress,
        token_name: str,
        token_ticker: str,
        num_decimals: int,
        can_freeze: bool,
        can_wipe: bool,
        can_pause: bool,
        can_transfer_nft_create_role: bool,
        can_change_owner: bool,
        can_upgrade: bool,
        can_add_special_roles: bool
    ) -> Transaction:
        self._notify_about_unsetting_burn_role_globally()

        parts: List[str] = [
            "registerMetaDCDT",
            arg_to_string(token_name),
            arg_to_string(token_ticker),
            arg_to_string(num_decimals),
            *([arg_to_string("canFreeze"), self._true_as_hex if can_freeze else self._false_as_hex]),
            *([arg_to_string("canWipe"), self._true_as_hex if can_wipe else self._false_as_hex]),
            *([arg_to_string("canPause"), self._true_as_hex if can_pause else self._false_as_hex]),
            *([arg_to_string("canTransferNFTCreateRole"), self._true_as_hex if can_transfer_nft_create_role else self._false_as_hex]),
            *([arg_to_string("canChangeOwner"), self._true_as_hex if can_change_owner else self._false_as_hex]),
            *([arg_to_string("canUpgrade"), self._true_as_hex if can_upgrade else self._false_as_hex]),
            *([arg_to_string("canAddSpecialRoles"), self._true_as_hex if can_add_special_roles else self._false_as_hex])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=self._config.issue_cost,
            gas_limit=self._config.gas_limit_issue,
            add_data_movement_gas=True,
            data_parts=parts,
        ).build()

    def create_transaction_for_registering_and_setting_roles(
        self,
        sender: IAddress,
        token_name: str,
        token_ticker: str,
        token_type: RegisterAndSetAllRolesTokenType,
        num_decimals: int
    ) -> Transaction:
        self._notify_about_unsetting_burn_role_globally()

        parts: List[str] = [
            "registerAndSetAllRoles",
            arg_to_string(token_name),
            arg_to_string(token_ticker),
            arg_to_string(token_type.value),
            arg_to_string(num_decimals)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=self._config.issue_cost,
            gas_limit=self._config.gas_limit_issue,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_setting_burn_role_globally(
        self,
        sender: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "setBurnRoleGlobally",
            arg_to_string(token_identifier)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_toggle_burn_role_globally,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_unsetting_burn_role_globally(
        self,
        sender: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "unsetBurnRoleGlobally",
            arg_to_string(token_identifier)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_toggle_burn_role_globally,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_setting_special_role_on_fungible_token(
        self,
        sender: IAddress,
        user: IAddress,
        token_identifier: str,
        add_role_local_mint: bool,
        add_role_local_burn: bool
    ) -> Transaction:
        parts: List[str] = [
            "setSpecialRole",
            arg_to_string(token_identifier),
            user.to_hex(),
            *([arg_to_string("DCDTRoleLocalMint")] if add_role_local_mint else []),
            *([arg_to_string("DCDTRoleLocalBurn")] if add_role_local_burn else [])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_set_special_role,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_setting_special_role_on_semi_fungible_token(
        self,
        sender: IAddress,
        user: IAddress,
        token_identifier: str,
        add_role_nft_create: bool,
        add_role_nft_burn: bool,
        add_role_nft_add_quantity: bool,
        add_role_dcdt_transfer_role: bool
    ) -> Transaction:
        parts: List[str] = [
            "setSpecialRole",
            arg_to_string(token_identifier),
            user.to_hex(),
            *([arg_to_string("DCDTRoleNFTCreate")] if add_role_nft_create else []),
            *([arg_to_string("DCDTRoleNFTBurn")] if add_role_nft_burn else []),
            *([arg_to_string("DCDTRoleNFTAddQuantity")] if add_role_nft_add_quantity else []),
            *([arg_to_string("DCDTTransferRole")] if add_role_dcdt_transfer_role else [])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_set_special_role,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_setting_special_role_on_non_fungible_token(
        self,
        sender: IAddress,
        user: IAddress,
        token_identifier: str,
        add_role_nft_create: bool,
        add_role_nft_burn: bool,
        add_role_nft_update_attributes: bool,
        add_role_nft_add_uri: bool,
        add_role_dcdt_transfer_role: bool
    ) -> Transaction:
        parts: List[str] = [
            "setSpecialRole",
            arg_to_string(token_identifier),
            user.to_hex(),
            *([arg_to_string("DCDTRoleNFTCreate")] if add_role_nft_create else []),
            *([arg_to_string("DCDTRoleNFTBurn")] if add_role_nft_burn else []),
            *([arg_to_string("DCDTRoleNFTUpdateAttributes")] if add_role_nft_update_attributes else []),
            *([arg_to_string("DCDTRoleNFTAddURI")] if add_role_nft_add_uri else []),
            *([arg_to_string("DCDTTransferRole")] if add_role_dcdt_transfer_role else [])
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_set_special_role,
            add_data_movement_gas=True,
            data_parts=parts,
        ).build()

    def create_transaction_for_creating_nft(
        self,
        sender: IAddress,
        token_identifier: str,
        initial_quantity: int,
        name: str,
        royalties: int,
        hash: str,
        attributes: bytes,
        uris: List[str]
    ) -> Transaction:
        parts: List[str] = [
            "DCDTNFTCreate",
            arg_to_string(token_identifier),
            arg_to_string(initial_quantity),
            arg_to_string(name),
            arg_to_string(royalties),
            arg_to_string(hash),
            arg_to_string(attributes),
            *map(arg_to_string, uris)
        ]

        # Note that the following is an approximation (a reasonable one):
        nft_data = name + hash + attributes.hex() + "".join(uris)
        storage_gas_limit = len(nft_data) * self._config.gas_limit_store_per_byte

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=sender,
            amount=None,
            gas_limit=self._config.gas_limit_dcdt_nft_create + storage_gas_limit,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_pausing(
        self,
        sender: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "pause",
            arg_to_string(token_identifier)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_pausing,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_unpausing(
        self,
        sender: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "unPause",
            arg_to_string(token_identifier)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_pausing,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_freezing(
        self,
        sender: IAddress,
        user: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "freeze",
            arg_to_string(token_identifier),
            user.to_hex()
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_freezing,
            add_data_movement_gas=True,
            data_parts=parts,
        ).build()

    def create_transaction_for_unfreezing(
        self,
        sender: IAddress,
        user: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "unFreeze",
            arg_to_string(token_identifier),
            user.to_hex()
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_freezing,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_wiping(
        self,
        sender: IAddress,
        user: IAddress,
        token_identifier: str
    ) -> Transaction:
        parts: List[str] = [
            "wipe",
            arg_to_string(token_identifier),
            user.to_hex()
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=self._config.dcdt_contract_address,
            amount=None,
            gas_limit=self._config.gas_limit_wiping,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_local_minting(
        self,
        sender: IAddress,
        token_identifier: str,
        supply_to_mint: int
    ) -> Transaction:
        parts: List[str] = [
            "DCDTLocalMint",
            arg_to_string(token_identifier),
            arg_to_string(supply_to_mint)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=sender,
            amount=None,
            gas_limit=self._config.gas_limit_dcdt_local_mint,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_local_burning(
        self,
        sender: IAddress,
        token_identifier: str,
        supply_to_burn: int
    ) -> Transaction:
        parts: List[str] = [
            "DCDTLocalBurn",
            arg_to_string(token_identifier),
            arg_to_string(supply_to_burn)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=sender,
            amount=None,
            gas_limit=self._config.gas_limit_dcdt_local_burn,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_updating_attributes(
        self,
        sender: IAddress,
        token_identifier: str,
        token_nonce: int,
        attributes: bytes
    ) -> Transaction:
        parts: List[str] = [
            "DCDTNFTUpdateAttributes",
            arg_to_string(token_identifier),
            arg_to_string(token_nonce),
            arg_to_string(attributes)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=sender,
            amount=None,
            gas_limit=self._config.gas_limit_dcdt_nft_update_attributes,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_adding_quantity(
        self,
        sender: IAddress,
        token_identifier: str,
        token_nonce: int,
        quantity_to_add: int
    ) -> Transaction:
        parts: List[str] = [
            "DCDTNFTAddQuantity",
            arg_to_string(token_identifier),
            arg_to_string(token_nonce),
            arg_to_string(quantity_to_add)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=sender,
            amount=None,
            gas_limit=self._config.gas_limit_dcdt_nft_add_quantity,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()

    def create_transaction_for_burning_quantity(
        self,
        sender: IAddress,
        token_identifier: str,
        token_nonce: int,
        quantity_to_burn: int
    ) -> Transaction:
        parts: List[str] = [
            "DCDTNFTBurn",
            arg_to_string(token_identifier),
            arg_to_string(token_nonce),
            arg_to_string(quantity_to_burn)
        ]

        return TransactionBuilder(
            config=self._config,
            sender=sender,
            receiver=sender,
            amount=None,
            gas_limit=self._config.gas_limit_dcdt_nft_burn,
            add_data_movement_gas=True,
            data_parts=parts
        ).build()



from pathlib import Path
from typing import Any, List, Protocol, Sequence, Union

from drt_sdk_core import Transaction
from drt_sdk_core.address import Address
from drt_sdk_core.code_metadata import CodeMetadata
from drt_sdk_core.constants import (CONTRACT_DEPLOY_ADDRESS,
                                           VM_TYPE_WASM_VM)
from drt_sdk_core.errors import BadUsageError
from drt_sdk_core.interfaces import IAddress, IToken, ITokenTransfer
from drt_sdk_core.serializer import arg_to_string, args_to_strings
from drt_sdk_core.transaction_factories.token_transfers_data_builder import \
    TokenTransfersDataBuilder
from drt_sdk_core.transaction_factories.transaction_builder import \
    TransactionBuilder


class IConfig(Protocol):
    chain_id: str
    min_gas_limit: int
    gas_limit_per_byte: int


class ITokenComputer(Protocol):
    def is_fungible(self, token: IToken) -> bool:
        ...

    def extract_identifier_from_extended_identifier(self, identifier: str) -> str:
        ...


class SmartContractTransactionsFactory:
    def __init__(self, config: IConfig, token_computer: ITokenComputer) -> None:
        self.config = config
        self.token_computer = token_computer
        self._data_args_builder = TokenTransfersDataBuilder(token_computer)

    def create_transaction_for_deploy(self,
                                      sender: IAddress,
                                      bytecode: Union[Path, bytes],
                                      gas_limit: int,
                                      arguments: Sequence[Any] = [],
                                      native_transfer_amount: int = 0,
                                      is_upgradeable: bool = True,
                                      is_readable: bool = True,
                                      is_payable: bool = False,
                                      is_payable_by_sc: bool = True) -> Transaction:
        if isinstance(bytecode, Path):
            bytecode = bytecode.read_bytes()

        metadata = CodeMetadata(is_upgradeable, is_readable, is_payable, is_payable_by_sc)

        parts = [
            arg_to_string(bytecode),
            arg_to_string(VM_TYPE_WASM_VM),
            str(metadata)
        ]

        parts += args_to_strings(arguments)

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=Address.new_from_bech32(CONTRACT_DEPLOY_ADDRESS),
            data_parts=parts,
            gas_limit=gas_limit,
            add_data_movement_gas=False,
            amount=native_transfer_amount
        ).build()

        return transaction

    def create_transaction_for_execute(self,
                                       sender: IAddress,
                                       contract: IAddress,
                                       function: str,
                                       gas_limit: int,
                                       arguments: Sequence[Any] = [],
                                       native_transfer_amount: int = 0,
                                       token_transfers: Sequence[ITokenTransfer] = []) -> Transaction:
        number_of_tokens = len(token_transfers)
        receiver = contract

        if native_transfer_amount and number_of_tokens:
            raise BadUsageError("Can't send both native token and custom tokens(DCDT/NFT)")

        data_parts: List[str] = []

        if len(token_transfers) == 1:
            transfer = token_transfers[0]

            if self.token_computer.is_fungible(transfer.token):
                data_parts = self._data_args_builder.build_args_for_dcdt_transfer(transfer=transfer)
            else:
                data_parts = self._data_args_builder.build_args_for_single_dcdt_nft_transfer(
                    transfer=transfer, receiver=receiver)
                receiver = sender
        elif len(token_transfers) > 1:
            data_parts = self._data_args_builder.build_args_for_multi_dcdt_nft_transfer(
                receiver=receiver, transfers=token_transfers)
            receiver = sender

        data_parts.append(function) if not data_parts else data_parts.append(arg_to_string(function))
        data_parts.extend(args_to_strings(arguments))

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=receiver,
            data_parts=data_parts,
            gas_limit=gas_limit,
            add_data_movement_gas=False,
            amount=native_transfer_amount
        ).build()

        return transaction

    def create_transaction_for_upgrade(self,
                                       sender: IAddress,
                                       contract: IAddress,
                                       bytecode: Union[Path, bytes],
                                       gas_limit: int,
                                       arguments: Sequence[Any] = [],
                                       native_transfer_amount: int = 0,
                                       is_upgradeable: bool = True,
                                       is_readable: bool = True,
                                       is_payable: bool = False,
                                       is_payable_by_sc: bool = True) -> Transaction:
        if isinstance(bytecode, Path):
            bytecode = bytecode.read_bytes()

        metadata = CodeMetadata(is_upgradeable, is_readable, is_payable, is_payable_by_sc)

        parts = [
            "upgradeContract",
            arg_to_string(bytecode),
            str(metadata)
        ]

        parts += args_to_strings(arguments)

        intent = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=contract,
            data_parts=parts,
            gas_limit=gas_limit,
            add_data_movement_gas=False,
            amount=native_transfer_amount
        ).build()

        return intent



from typing import List, Optional, Protocol, Sequence

from drt_sdk_core.interfaces import (IAddress, IGasLimit, IGasPrice,
                                            INonce, ITokenPayment,
                                            ITransactionValue)
from drt_sdk_core.serializer import arg_to_string
from drt_sdk_core.transaction_builders.transaction_builder import (
    ITransactionBuilderConfiguration, TransactionBuilder)


class IDCDTTransferConfiguration(ITransactionBuilderConfiguration, Protocol):
    gas_limit_dcdt_transfer: IGasLimit
    additional_gas_for_dcdt_transfer: IGasLimit


class IDCDTNFTTransferConfiguration(ITransactionBuilderConfiguration, Protocol):
    gas_limit_dcdt_nft_transfer: IGasLimit
    additional_gas_for_dcdt_nft_transfer: IGasLimit


class REWATransferBuilder(TransactionBuilder):
    def __init__(self,
                 config: ITransactionBuilderConfiguration,
                 sender: IAddress,
                 receiver: IAddress,
                 payment: ITokenPayment,
                 nonce: Optional[INonce] = None,
                 data: Optional[str] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        assert payment.is_rewa()
        super().__init__(config, nonce, payment.amount_as_integer, gas_limit, gas_price)
        self.sender = sender
        self.receiver = receiver
        self.data = data

    def _estimate_execution_gas(self) -> IGasLimit:
        return 0

    def _build_payload_parts(self) -> List[str]:
        return [self.data] if self.data else []


class DCDTTransferBuilder(TransactionBuilder):
    def __init__(self,
                 config: IDCDTTransferConfiguration,
                 sender: IAddress,
                 receiver: IAddress,
                 payment: ITokenPayment,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.gas_limit_dcdt_transfer = config.gas_limit_dcdt_transfer
        self.additional_gas_for_dcdt_transfer = config.additional_gas_for_dcdt_transfer

        self.sender = sender
        self.receiver = receiver
        self.payment = payment

    def _estimate_execution_gas(self) -> IGasLimit:
        return self.gas_limit_dcdt_transfer + self.additional_gas_for_dcdt_transfer

    def _build_payload_parts(self) -> List[str]:
        return [
            "DCDTTransfer",
            arg_to_string(self.payment.token_identifier),
            arg_to_string(self.payment.amount_as_integer)
        ]


class DCDTNFTTransferBuilder(TransactionBuilder):
    def __init__(self,
                 config: IDCDTNFTTransferConfiguration,
                 sender: IAddress,
                 destination: IAddress,
                 payment: ITokenPayment,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.gas_limit_dcdt_nft_transfer = config.gas_limit_dcdt_nft_transfer
        self.additional_gas_for_dcdt_nft_transfer = config.additional_gas_for_dcdt_nft_transfer

        self.sender = sender
        self.receiver = sender
        self.destination = destination
        self.payment = payment

    def _estimate_execution_gas(self) -> IGasLimit:
        return self.gas_limit_dcdt_nft_transfer + self.additional_gas_for_dcdt_nft_transfer

    def _build_payload_parts(self) -> List[str]:
        return [
            "DCDTNFTTransfer",
            arg_to_string(self.payment.token_identifier),
            arg_to_string(self.payment.token_nonce),
            arg_to_string(self.payment.amount_as_integer),
            arg_to_string(self.destination)
        ]


class MultiDCDTNFTTransferBuilder(TransactionBuilder):
    def __init__(self,
                 config: IDCDTNFTTransferConfiguration,
                 sender: IAddress,
                 destination: IAddress,
                 payments: Sequence[ITokenPayment],
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.gas_limit_dcdt_nft_transfer = config.gas_limit_dcdt_nft_transfer
        self.additional_gas_for_dcdt_nft_transfer = config.additional_gas_for_dcdt_nft_transfer

        self.sender = sender
        self.receiver = sender
        self.destination = destination
        self.payments = payments

    def _estimate_execution_gas(self) -> IGasLimit:
        return (self.gas_limit_dcdt_nft_transfer + self.additional_gas_for_dcdt_nft_transfer) * len(self.payments)

    def _build_payload_parts(self) -> List[str]:
        parts = [
            "MultiDCDTNFTTransfer",
            arg_to_string(self.destination),
            arg_to_string(len(self.payments))
        ]

        for payment in self.payments:
            parts.extend([
                arg_to_string(payment.token_identifier),
                arg_to_string(payment.token_nonce),
                arg_to_string(payment.amount_as_integer)
            ])

        return parts




from drt_sdk_core.transaction_builders.contract_builders import (
    ContractCallBuilder, ContractDeploymentBuilder, ContractUpgradeBuilder)
from drt_sdk_core.transaction_builders.default_configuration import \
    DefaultTransactionBuildersConfiguration
from drt_sdk_core.transaction_builders.dcdt_builders import \
    DCDTIssueBuilder
from drt_sdk_core.transaction_builders.relayed_v1_builder import \
    RelayedTransactionV1Builder
from drt_sdk_core.transaction_builders.relayed_v2_builder import \
    RelayedTransactionV2Builder
from drt_sdk_core.transaction_builders.transaction_builder import \
    TransactionBuilder
from drt_sdk_core.transaction_builders.transfers_builders import (
    REWATransferBuilder, DCDTNFTTransferBuilder, DCDTTransferBuilder,
    MultiDCDTNFTTransferBuilder)

__all__ = [
    "TransactionBuilder",
    "DefaultTransactionBuildersConfiguration",
    "ContractCallBuilder", "ContractDeploymentBuilder", "ContractUpgradeBuilder",
    "REWATransferBuilder", "DCDTNFTTransferBuilder", "DCDTTransferBuilder", "MultiDCDTNFTTransferBuilder",
    "DCDTIssueBuilder", "RelayedTransactionV1Builder", "RelayedTransactionV2Builder"
]



import pytest

from drt_sdk_core.address import Address
from drt_sdk_core.errors import (ErrInvalidGasLimitForInnerTransaction,
                                        ErrInvalidRelayerV2BuilderArguments)
from drt_sdk_core.testutils.wallets import load_wallets
from drt_sdk_core.transaction import Transaction, TransactionComputer
from drt_sdk_core.transaction_builders.relayed_v2_builder import \
    RelayedTransactionV2Builder


class NetworkConfig:
    def __init__(self) -> None:
        self.min_gas_limit = 50_000
        self.gas_per_data_byte = 1_500
        self.gas_price_modifier = 0.01
        self.chain_id = "T"


class TestRelayedV2Builder:
    wallets = load_wallets()
    alice = wallets["alice"]
    bob = wallets["bob"]

    def test_without_arguments(self):
        relayed_builder = RelayedTransactionV2Builder()

        with pytest.raises(ErrInvalidRelayerV2BuilderArguments):
            relayed_builder.build()

    def test_with_inner_tx_gas_limit(self):
        builder = RelayedTransactionV2Builder()
        network_config = NetworkConfig()
        transaction_computer = TransactionComputer()

        inner_tx = Transaction(
            chain_id=network_config.chain_id,
            sender=self.alice.label,
            receiver="moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv",
            gas_limit=10,
            nonce=15,
            data=b"getContractConfig"
        )
        inner_tx.signature = self.alice.secret_key.sign(transaction_computer.compute_bytes_for_signing(inner_tx))

        builder.set_network_config(network_config)
        builder.set_inner_transaction_gas_limit(10)
        builder.set_inner_transaction(inner_tx)
        builder.set_relayer_address(Address.new_from_bech32(self.alice.label))

        with pytest.raises(ErrInvalidGasLimitForInnerTransaction):
            builder.build()

    def test_compute_relayed_v2_transaction(self):
        network_config = NetworkConfig()
        transaction_computer = TransactionComputer()

        inner_tx = Transaction(
            chain_id=network_config.chain_id,
            sender=self.bob.label,
            receiver="moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv",
            gas_limit=0,
            nonce=15,
            data=b"getContractConfig"
        )
        inner_tx.signature = self.bob.secret_key.sign(transaction_computer.compute_bytes_for_signing(inner_tx))

        builder = RelayedTransactionV2Builder()
        builder.set_inner_transaction(inner_tx)
        builder.set_inner_transaction_gas_limit(60_000_000)
        builder.set_relayer_nonce(37)
        builder.set_network_config(network_config)
        builder.set_relayer_address(Address.new_from_bech32(self.alice.label))

        relayed_tx = builder.build()

        relayed_tx.sender = self.alice.label
        relayed_tx.signature = self.alice.secret_key.sign(transaction_computer.compute_bytes_for_signing(relayed_tx))

        assert relayed_tx.version == 2
        assert relayed_tx.options == 0
        assert relayed_tx.nonce == 37
        assert relayed_tx.data.decode() == "relayedTxV2@000000000000000000010000000000000000000000000000000000000002ffff@0f@676574436f6e7472616374436f6e666967@9b5f2023b8423891f994f7c2fe1560ef9a1370e13243acabe3a457320416493323ba1f5e7e1dab9932ec8bab18b60745c3a283f737fc42aa06b12db92200eb00"



from dataclasses import dataclass

from drt_sdk_core.address import Address
from drt_sdk_core.interfaces import (IAddress, IChainID, IGasPrice,
                                            ITransactionValue)


@dataclass
class DefaultTransactionBuildersConfiguration:
    chain_id: IChainID
    min_gas_price: IGasPrice = 1000000000
    min_gas_limit = 50000
    gas_limit_per_byte = 1500

    issue_cost: ITransactionValue = 50000000000000000
    gas_limit_dcdt_issue = 60000000
    gas_limit_dcdt_transfer = 200000
    gas_limit_dcdt_nft_transfer = 200000
    additional_gas_for_dcdt_transfer = 100000
    additional_gas_for_dcdt_nft_transfer = 800000

    dcdt_contract_address: IAddress = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv")
    deployment_address: IAddress = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqhsx6tv")



import pytest

from drt_sdk_core.errors import BadUsageError
from drt_sdk_core.tokens import (Token, TokenComputer,
                                        TokenIdentifierParts)


class TestTokenComputer:
    token_computer = TokenComputer()

    def test_token_is_fungible(self):
        fungible_token = Token("TEST-123456")
        non_fungible_token = Token("NFT-987654", 7)

        assert self.token_computer.is_fungible(fungible_token)
        assert not self.token_computer.is_fungible(non_fungible_token)

    def test_extract_nonce_from_extended_identifier(self):
        extended_nft_identifier = "TEST-123456-0a"
        extended_fungible_identifier = "FNG-123456"

        assert self.token_computer.extract_nonce_from_extended_identifier(extended_nft_identifier) == 10
        assert self.token_computer.extract_nonce_from_extended_identifier(extended_fungible_identifier) == 0

    def test_extract_identifier_from_extended_identifier(self):
        extended_nft_identifier = "TEST-123456-0a"
        extended_fungible_identifier = "FNG-123456"

        assert self.token_computer.extract_identifier_from_extended_identifier(extended_nft_identifier) == "TEST-123456"
        assert self.token_computer.extract_identifier_from_extended_identifier(extended_fungible_identifier) == "FNG-123456"

    def test_extract_ticker_from_identifier(self):
        fungible_identifier = "FNG-123456"
        non_fungible_identifier = "NFT-987654-0a"

        assert self.token_computer.extract_ticker_from_identifier(fungible_identifier) == "FNG"
        assert self.token_computer.extract_ticker_from_identifier(non_fungible_identifier) == "NFT"

    def test_parse_extended_identifier_parts(self):
        fungible_identifier = "FNG-123456"
        non_fungible_identifier = "NFT-987654-0a"

        fungible_parts = self.token_computer.parse_extended_identifier_parts(fungible_identifier)
        non_fungible_parts = self.token_computer.parse_extended_identifier_parts(non_fungible_identifier)

        assert fungible_parts.ticker == "FNG"
        assert fungible_parts.random_sequence == "123456"
        assert fungible_parts.nonce == 0

        assert non_fungible_parts.ticker == "NFT"
        assert non_fungible_parts.random_sequence == "987654"
        assert non_fungible_parts.nonce == 10

    def test_compute_extended_identifier_from_identifier_and_bad_nonce(self):
        fungible_identifier = "FNG-123456"
        fungible_nonce = -10

        with pytest.raises(BadUsageError, match="The token nonce can not be less than 0"):
            self.token_computer.compute_extended_identifier_from_identifier_and_nonce(
                fungible_identifier, fungible_nonce
            )

    def test_compute_extended_identifier_from_identifier_and_nonce(self):
        fungible_identifier = "FNG-123456"
        fungible_nonce = 0

        non_fungible_identifier = "NFT-987654"
        non_fungible_nonce = 10

        fungible_token_identifier = self.token_computer.compute_extended_identifier_from_identifier_and_nonce(
            fungible_identifier, fungible_nonce
        )
        nft_identifier = self.token_computer.compute_extended_identifier_from_identifier_and_nonce(
            non_fungible_identifier, non_fungible_nonce
        )

        assert fungible_token_identifier == "FNG-123456"
        assert nft_identifier == "NFT-987654-0a"

    def test_compute_extended_identifier_from_parts(self):
        fungible_parts = TokenIdentifierParts("FNG", "123456", 0)
        nft_parts = TokenIdentifierParts("NFT", "987654", 10)

        fungible_identifier = self.token_computer.compute_extended_identifier_from_parts(fungible_parts)
        nft_identifier = self.token_computer.compute_extended_identifier_from_parts(nft_parts)

        assert fungible_identifier == "FNG-123456"
        assert nft_identifier == "NFT-987654-0a"



from drt_sdk_core.address import Address
from drt_sdk_core.contract_query_builder import ContractQueryBuilder


def test_contract_query_builder():
    contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgquzmh78klkqwt0p4rjys0qtp3la07gz4d396q7vfu0t")
    caller = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")

    builder = ContractQueryBuilder(
        contract=contract,
        function="getFoobar",
        call_arguments=[42, "test", -836623209073744937290891644],
        caller=caller,
        value=1
    )

    query = builder.build()

    assert query.contract == contract
    assert query.function == "getFoobar"
    assert query.encoded_arguments == ["2a", "74657374", "fd4bf624f561bc2894c6fe84"]
    assert query.caller == caller
    assert query.value == 1



from typing import List, Optional, Protocol

from drt_sdk_core.constants import ARGS_SEPARATOR
from drt_sdk_core.interfaces import IAddress
from drt_sdk_core.transaction import Transaction


class IConfig(Protocol):
    chain_id: str
    min_gas_limit: int
    gas_limit_per_byte: int


class TransactionBuilder:
    def __init__(self,
                 config: IConfig,
                 sender: IAddress,
                 receiver: IAddress,
                 data_parts: List[str],
                 gas_limit: int,
                 add_data_movement_gas: bool,
                 amount: Optional[int] = None) -> None:
        self.config = config
        self.sender = sender
        self.receiver = receiver
        self.data_parts = data_parts
        self.provided_gas_limit = gas_limit
        self.add_data_movement_gas = add_data_movement_gas
        self.amount = amount

    def compute_gas_limit(self, payload: bytes) -> int:
        if not self.add_data_movement_gas:
            return self.provided_gas_limit

        data_movement_gas = self.config.min_gas_limit + self.config.gas_limit_per_byte * len(payload)
        gas = data_movement_gas + self.provided_gas_limit

        return gas

    def build_transaction_payload(self, parts: List[str]) -> bytes:
        data = ARGS_SEPARATOR.join(parts)
        return data.encode("utf-8")

    def build(self) -> Transaction:
        data = self.build_transaction_payload(self.data_parts)
        gas_limit = self.compute_gas_limit(data)

        transaction = Transaction(
            sender=self.sender.to_bech32(),
            receiver=self.receiver.to_bech32(),
            gas_limit=gas_limit,
            chain_id=self.config.chain_id,
            data=data,
            value=self.amount
        )

        return transaction



from Cryptodome.Hash import keccak

from drt_sdk_core.interfaces import IMessage


class Message:
    def __init__(self, data: bytes, signature: bytes = b"") -> None:
        self.data = data
        self.signature = signature


class MessageComputer:
    """
    Also see: 
     - https://github.com/DharitriOne/drt-sdk-js-core/blob/v11.2.0/src/signableMessage.ts
     - https://eips.ethereum.org/EIPS/eip-712 (in the past, it served as a basis for the implementation)
    """

    def __init__(self) -> None:
        pass

    def compute_bytes_for_signing(self, message: IMessage) -> bytes:
        PREFIX = bytes.fromhex("17456c726f6e64205369676e6564204d6573736167653a0a")
        size = str(len(message.data)).encode()
        content = PREFIX + size + message.data
        content_hash = keccak.new(digest_bits=256).update(content).digest()

        return content_hash



class SaveKeyValuesBuilder:
    def __init__(self) -> None:
        raise NotImplementedError()


class SetUsernameBuilder:
    def __init__(self) -> None:
        raise NotImplementedError()


class ChangeOwnerAddressBuilder:
    def __init__(self) -> None:
        raise NotImplementedError()


class ClaimDeveloperRewardsBuilder:
    def __init__(self) -> None:
        raise NotImplementedError()



from drt_sdk_core.transaction_builders.default_configuration import \
    DefaultTransactionBuildersConfiguration


def test_default_configuration():
    config = DefaultTransactionBuildersConfiguration(chain_id="D")

    assert config.chain_id == "D"
    assert config.min_gas_price == 1000000000
    assert config.min_gas_limit == 50000
    assert config.gas_limit_per_byte == 1500



from drt_sdk_core.transaction_factories.delegation_transactions_factory import \
    DelegationTransactionsFactory
from drt_sdk_core.transaction_factories.smart_contract_transactions_factory import \
    SmartContractTransactionsFactory
from drt_sdk_core.transaction_factories.token_management_transactions_factory import (
    RegisterAndSetAllRolesTokenType, TokenManagementTransactionsFactory)
from drt_sdk_core.transaction_factories.transactions_factory_config import \
    TransactionsFactoryConfig
from drt_sdk_core.transaction_factories.transfer_transactions_factory import \
    TransferTransactionsFactory

__all__ = [
    "DelegationTransactionsFactory",
    "TokenManagementTransactionsFactory",
    "RegisterAndSetAllRolesTokenType",
    "TransactionsFactoryConfig",
    "SmartContractTransactionsFactory",
    "TransferTransactionsFactory"
]




from typing import Protocol

from drt_sdk_core import Address
from drt_sdk_core.codec import decode_unsigned_number
from drt_sdk_core.transaction_parsers.token_operations_outcome_parser_types import (
    AddQuantityOutcome, BurnOutcome, BurnQuantityOutcome, DCDTIssueOutcome,
    FreezingOutcome, MintOutcome, NFTCreateOutcome, PausingOutcome,
    RegisterAndSetAllRolesOutcome, SetSpecialRoleOutcome,
    UpdateAttributesOutcome, WipingOutcome)
from drt_sdk_core.transaction_parsers.transaction_on_network_wrapper import (
    ITransactionEvent, ITransactionOnNetwork, TransactionOnNetworkWrapper)


class IConfig(Protocol):
    address_hrp: str


class TokenOperationsOutcomeParser:
    def __init__(self, config: IConfig) -> None:
        self._config = config

    def parse_issue_fungible(self, transaction: ITransactionOnNetwork) -> DCDTIssueOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("issue")
        token_identifier = self._extract_token_identifier(event)
        return DCDTIssueOutcome(token_identifier)

    def parse_issue_non_fungible(self, transaction: ITransactionOnNetwork) -> DCDTIssueOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("issueNonFungible")
        token_identifier = self._extract_token_identifier(event)
        return DCDTIssueOutcome(token_identifier)

    def parse_issue_semi_fungible(self, transaction: ITransactionOnNetwork) -> DCDTIssueOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("issueSemiFungible")
        token_identifier = self._extract_token_identifier(event)
        return DCDTIssueOutcome(token_identifier)

    def parse_register_meta_dcdt(self, transaction: ITransactionOnNetwork) -> DCDTIssueOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("registerMetaDCDT")
        token_identifier = self._extract_token_identifier(event)
        return DCDTIssueOutcome(token_identifier)

    def parse_register_and_set_all_roles(self, transaction: ITransactionOnNetwork) -> RegisterAndSetAllRolesOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event_register = wrapper.find_single_event_by_identifier("registerAndSetAllRoles")
        token_identifier = self._extract_token_identifier(event_register)

        event_set_role = wrapper.find_single_event_by_identifier("DCDTSetRole")
        roles = [str(topic) for topic in event_set_role.topics[3:]]

        return RegisterAndSetAllRolesOutcome(token_identifier, roles)

    def parse_set_burn_role_globally(self, transaction: ITransactionOnNetwork) -> None:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

    def parse_unset_burn_role_globally(self, transaction: ITransactionOnNetwork) -> None:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

    def parse_set_special_role(self, transaction: ITransactionOnNetwork) -> SetSpecialRoleOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTSetRole")
        user_address = event.address.to_bech32()
        token_identifier = self._extract_token_identifier(event)
        roles = [str(topic) for topic in event.topics[3:]]

        return SetSpecialRoleOutcome(user_address, token_identifier, roles)

    def parse_nft_create(self, transaction: ITransactionOnNetwork) -> NFTCreateOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTNFTCreate")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        initial_quantity = self._extract_amount(event)

        return NFTCreateOutcome(token_identifier, nonce, initial_quantity)

    def parse_local_mint(self, transaction: ITransactionOnNetwork) -> MintOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTLocalMint")
        user_address = event.address.to_bech32()
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        minted_supply = self._extract_amount(event)

        return MintOutcome(user_address, token_identifier, nonce, minted_supply)

    def parse_local_burn(self, transaction: ITransactionOnNetwork) -> BurnOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTLocalBurn")
        user_address = event.address.to_bech32()
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        burnt_supply = self._extract_amount(event)

        return BurnOutcome(user_address, token_identifier, nonce, burnt_supply)

    def parse_pause(self, transaction: ITransactionOnNetwork) -> PausingOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        _ = wrapper.find_single_event_by_identifier("DCDTPause")
        return PausingOutcome()

    def parse_unpause(self, transaction: ITransactionOnNetwork) -> PausingOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        _ = wrapper.find_single_event_by_identifier("DCDTUnPause")
        return PausingOutcome()

    def parse_freeze(self, transaction: ITransactionOnNetwork) -> FreezingOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTFreeze")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        balance = self._extract_amount(event)
        user_address = self._extract_address(event)

        return FreezingOutcome(user_address, token_identifier, nonce, balance)

    def parse_unfreeze(self, transaction: ITransactionOnNetwork) -> FreezingOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTUnFreeze")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        balance = self._extract_amount(event)
        user_address = self._extract_address(event)

        return FreezingOutcome(user_address, token_identifier, nonce, balance)

    def parse_wipe(self, transaction: ITransactionOnNetwork) -> WipingOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTWipe")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        balance = self._extract_amount(event)
        user_address = self._extract_address(event)

        return WipingOutcome(user_address, token_identifier, nonce, balance)

    def parse_update_attributes(self, transaction: ITransactionOnNetwork) -> UpdateAttributesOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTNFTUpdateAttributes")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        attributes = event.topics[3].raw

        return UpdateAttributesOutcome(token_identifier, nonce, attributes)

    def parse_add_quantity(self, transaction: ITransactionOnNetwork) -> AddQuantityOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTNFTAddQuantity")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        added_quantity = self._extract_amount(event)

        return AddQuantityOutcome(token_identifier, nonce, added_quantity)

    def parse_burn_quantity(self, transaction: ITransactionOnNetwork) -> BurnQuantityOutcome:
        wrapper = TransactionOnNetworkWrapper.from_transaction(transaction)
        wrapper.ensure_no_error()

        event = wrapper.find_single_event_by_identifier("DCDTNFTBurn")
        token_identifier = self._extract_token_identifier(event)
        nonce = self._extract_nonce(event)
        burnt_quantity = self._extract_amount(event)

        return BurnQuantityOutcome(token_identifier, nonce, burnt_quantity)

    def _extract_token_identifier(self, event: ITransactionEvent) -> str:
        return str(event.topics[0])

    def _extract_nonce(self, event: ITransactionEvent) -> int:
        return decode_unsigned_number(event.topics[1].raw)

    def _extract_amount(self, event: ITransactionEvent) -> int:
        return decode_unsigned_number(event.topics[2].raw)

    def _extract_address(self, event: ITransactionEvent) -> str:
        return Address(event.topics[3].raw, self._config.address_hrp).to_bech32()




from drt_sdk_core.address import Address
from drt_sdk_core.token_payment import TokenPayment
from drt_sdk_core.transaction_builders.default_configuration import \
    DefaultTransactionBuildersConfiguration
from drt_sdk_core.transaction_builders.transfers_builders import (
    REWATransferBuilder, DCDTNFTTransferBuilder, DCDTTransferBuilder,
    MultiDCDTNFTTransferBuilder)

dummyConfig = DefaultTransactionBuildersConfiguration(chain_id="D")


def test_rewa_transfer_builder():
    alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")
    payment = TokenPayment.rewa_from_amount("1.00")

    # With "data" field
    builder = REWATransferBuilder(
        config=dummyConfig,
        sender=alice,
        receiver=bob,
        payment=payment,
        data="for the book"
    )

    payload = builder.build_payload()
    tx = builder.build()
    assert payload.data == b"for the book"
    assert tx.chain_id == "D"
    assert tx.sender == alice.to_bech32()
    assert tx.receiver == bob.to_bech32()
    assert tx.gas_limit == 50000 + payload.length() * 1500
    assert tx.data.decode() == str(payload)

    # Without "data" field
    builder = REWATransferBuilder(
        config=dummyConfig,
        sender=alice,
        receiver=bob,
        payment=payment
    )

    payload = builder.build_payload()
    tx = builder.build()
    assert payload.data == b""
    assert tx.chain_id == "D"
    assert tx.sender == alice.to_bech32()
    assert tx.receiver == bob.to_bech32()
    assert tx.gas_limit == 50000
    assert tx.data.decode() == str(payload)


def test_dcdt_transfer_builder():
    alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")
    payment = TokenPayment.fungible_from_amount("COUNTER-8b028f", "100.00", 2)

    builder = DCDTTransferBuilder(
        config=dummyConfig,
        sender=alice,
        receiver=bob,
        payment=payment
    )

    payload = builder.build_payload()
    tx = builder.build()
    assert payload.data == b"DCDTTransfer@434f554e5445522d386230323866@2710"
    assert tx.chain_id == "D"
    assert tx.sender == alice.to_bech32()
    assert tx.receiver == bob.to_bech32()
    assert tx.gas_limit == 50000 + payload.length() * 1500 + 100000 + 200000
    assert tx.data.decode() == str(payload)


def test_dcdt_nft_transfer_builder():
    alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")
    payment = TokenPayment.non_fungible("TEST-38f249", 1)

    builder = DCDTNFTTransferBuilder(
        config=dummyConfig,
        sender=alice,
        destination=bob,
        payment=payment
    )

    payload = builder.build_payload()
    tx = builder.build()
    assert payload.data == b"DCDTNFTTransfer@544553542d333866323439@01@01@8049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f8"
    assert tx.chain_id == "D"
    assert tx.sender == alice.to_bech32()
    assert tx.receiver == alice.to_bech32()
    assert tx.gas_limit == 50000 + payload.length() * 1500 + 200000 + 800000
    assert tx.data.decode() == str(payload)


def test_multi_dcdt_nft_transfer_builder():
    alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")

    payment_one = TokenPayment.non_fungible("TEST-38f249", 1)
    payment_two = TokenPayment.fungible_from_amount("BAR-c80d29", "10.00", 18)

    builder = MultiDCDTNFTTransferBuilder(
        config=dummyConfig,
        sender=alice,
        destination=bob,
        payments=[payment_one, payment_two]
    )

    payload = builder.build_payload()
    tx = builder.build()
    assert payload.data == b"MultiDCDTNFTTransfer@8049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f8@02@544553542d333866323439@01@01@4241522d633830643239@@8ac7230489e80000"
    assert tx.chain_id == "D"
    assert tx.sender == alice.to_bech32()
    assert tx.receiver == alice.to_bech32()
    assert tx.gas_limit == 50000 + payload.length() * 1500 + 2 * (200000 + 800000)
    assert tx.data.decode() == str(payload)



import base64

from drt_sdk_core.interfaces import ITransactionPayload


class TransactionPayload(ITransactionPayload):
    def __init__(self, data: bytes) -> None:
        self.data = data

    @classmethod
    def empty(cls) -> 'TransactionPayload':
        return TransactionPayload(bytes())

    @classmethod
    def from_str(cls, data: str) -> 'TransactionPayload':
        """
        Creates a TransactionPayload from a utf-8 string.
        """
        return TransactionPayload(data.encode("utf-8"))

    @classmethod
    def from_encoded_str(cls, encoded: str) -> 'TransactionPayload':
        """
        Creates a TransactionPayload from a base-64 encoded string.
        """
        data = base64.b64decode(encoded)
        return TransactionPayload(data)

    def is_empty(self) -> bool:
        return len(self.data) == 0

    def length(self) -> int:
        return len(self.data)

    def encoded(self) -> str:
        return base64.b64encode(self.data).decode()

    def __str__(self) -> str:
        return self.data.decode("utf-8")



import pytest
from drt_sdk_wallet import UserSecretKey

from drt_sdk_core.errors import NotEnoughGasError
from drt_sdk_core.proto.transaction_serializer import ProtoSerializer
from drt_sdk_core.testutils.wallets import load_wallets
from drt_sdk_core.transaction import Transaction, TransactionComputer


class NetworkConfig:
    def __init__(self, min_gas_limit: int = 50000) -> None:
        self.min_gas_limit = min_gas_limit
        self.gas_per_data_byte = 1500
        self.gas_price_modifier = 0.01
        self.chain_id = "D"


class TestTransaction:
    wallets = load_wallets()
    alice = wallets["alice"]
    carol = wallets["carol"]
    transaction_computer = TransactionComputer()

    def test_serialize_for_signing(self):
        sender = "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        receiver = "moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk"

        transaction = Transaction(
            nonce=89,
            sender=sender,
            receiver=receiver,
            value=0,
            gas_limit=50000,
            gas_price=1000000000,
            chain_id="D",
            version=1
        )
        serialized_tx = self.transaction_computer.compute_bytes_for_signing(transaction)
        assert serialized_tx.decode() == r"""{"nonce":89,"value":"0","receiver":"moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk","sender":"moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8","gasPrice":1000000000,"gasLimit":50000,"chainID":"D","version":1}"""

        transaction = Transaction(
            nonce=90,
            sender=sender,
            receiver=receiver,
            value=1000000000000000000,
            data=b"hello",
            gas_limit=70000,
            gas_price=1000000000,
            chain_id="D",
            version=1
        )
        serialized_tx = self.transaction_computer.compute_bytes_for_signing(transaction)
        assert serialized_tx.decode() == r"""{"nonce":90,"value":"1000000000000000000","receiver":"moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk","sender":"moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8","gasPrice":1000000000,"gasLimit":70000,"data":"aGVsbG8=","chainID":"D","version":1}"""

    def test_with_usernames(self):
        transaction = Transaction(
            chain_id="T",
            sender=self.carol.label,
            receiver=self.alice.label,
            nonce=204,
            gas_limit=50000,
            sender_username="carol",
            receiver_username="alice",
            value=1000000000000000000
        )

        transaction.signature = self.carol.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))
        assert transaction.signature.hex() == "3f3f645f9ac38142c1087a76b855e48ed0aa44185146df5f3317b8887e7b59f195d6093d3c2b30fb693a27e3175a199ef04a6e650101ab3f4b3649f0fe2f5204"

    def test_compute_transaction_hash(self):
        transaction = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            gas_limit=100000,
            chain_id="D",
            nonce=17243,
            value=1000000000000,
            data=b"testtx",
            version=2,
            signature=bytes.fromhex("eaa9e4dfbd21695d9511e9754bde13e90c5cfb21748a339a79be11f744c71872e9fe8e73c6035c413f5f08eef09e5458e9ea6fc315ff4da0ab6d000b450b2a07")
        )
        tx_hash = self.transaction_computer.compute_transaction_hash(transaction)
        assert tx_hash.hex() == "169b76b752b220a76a93aeebc462a1192db1dc2ec9d17e6b4d7b0dcc91792f03"

    def test_compute_transaction_hash_with_usernames(self):
        transaction = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            gas_limit=100000,
            chain_id="D",
            nonce=17244,
            value=1000000000000,
            data=b"testtx",
            version=2,
            sender_username="alice",
            receiver_username="alice",
            signature=bytes.fromhex("807bcd7de5553ea6dfc57c0510e84d46813c5963d90fec50991c500091408fcf6216dca48dae16a579a1611ed8b2834bae8bd0027dc17eb557963f7151b82c07")
        )
        tx_hash = self.transaction_computer.compute_transaction_hash(transaction)
        assert tx_hash.hex() == "41b5acf7ebaf4a9165a64206b6ebc02021b3adda55ffb2a2698aac2e7004dc29"

    def test_compute_transaction_fee_insufficient(self):
        transaction = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            gas_limit=50000,
            chain_id="D",
            data=b"toolittlegaslimit",
        )

        with pytest.raises(NotEnoughGasError):
            self.transaction_computer.compute_transaction_fee(transaction, NetworkConfig())

    def test_compute_transaction_fee(self):
        transaction = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            gas_price=500,
            gas_limit=20,
            chain_id="D",
        )

        computed_gas = self.transaction_computer.compute_transaction_fee(transaction, NetworkConfig(min_gas_limit=10))
        assert computed_gas == 5050

    def test_compute_transaction_fee_with_data_field(self):
        transaction = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            gas_price=500,
            gas_limit=12010,
            chain_id="D",
            data=b"testdata"
        )

        computed_gas = self.transaction_computer.compute_transaction_fee(transaction, NetworkConfig(min_gas_limit=10))
        assert computed_gas == 6005000

    def test_compute_transaction_with_guardian_fields(self):

        sender_secret_key_hex = "3964a58b0debd802f67239c30aa2b3a75fff1842c203587cb590d03d20e32415"
        sender_secret_key = UserSecretKey(bytes.fromhex(sender_secret_key_hex))

        transaction = Transaction(
            sender="moa1fp4zaxvyc8jh99vauwns99kvs9tn0k6cwrr0zpyz2jvyurcepuhsy62spn",
            receiver="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            gas_limit=139000,
            gas_price=1000000000,
            chain_id="D",
            nonce=2,
            value=1000000000000000000,
            data=b"this is a test transaction",
            version=2,
            options=2,
            guardian="moa1nn8apn09vmf72l7kzr3nd90rr5r2q74he7hseghs3v68c5p7ud2q60gxe2",
            guardian_signature=bytes.fromhex("487150c26d38a01fe19fbe26dac20ec2b42ec3abf5763a47a508e62bcd6ad3437c4d404684442e864a1dbad446dc0f852889a09f0650b5fdb55f4ee18147920d")
        )

        transaction.signature = sender_secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))
        assert transaction.signature.hex() == "b77c33efa5cd196245049ecbe176afa4b293027552e0ed9dead3e2b1eb1c0f8a749351a5eb1b398ec09b30da6ec0c81ab833dbd188482490bb046a78133ca808"

        tx_hash = self.transaction_computer.compute_transaction_hash(transaction)
        assert tx_hash.hex() == "010347ada2669f563ad5c6d339afb5ccdb4cf299aee8af9ec7acca24acd006cf"

    # this test was done to mimic the one in drt-chain-go
    def test_compute_transaction_with_dummy_guardian(self):
        alice_private_key_hex = "413f42575f7f26fad3317a778771212fdb80245850981e48b58a4f25e344e8f9"
        alice_secret_key = UserSecretKey(bytes.fromhex(alice_private_key_hex))

        transaction = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk",
            gas_limit=150000,
            chain_id="local-testnet",
            gas_price=1000000000,
            data=b"test data field",
            version=2,
            options=2,
            nonce=92,
            value=123456789000000000000000000000,
            guardian="moa1x23lzn8483xs2su4fak0r0dqx6w38enpmmqf2yrkylwq7mfnvyhstcgmz5",
            guardian_signature=bytes([0] * 64)
        )

        transaction.signature = alice_secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))
        assert transaction.signature.hex() == "c74c3e5b276c32ae72ab1a0bac17939d7577e55e1467cd3b4d4a45b04ad3957e94a855657b42601384819502882d559dd5f8e31a93097e113fe9fa8261515104"

        proto_serializer = ProtoSerializer()
        serialized = proto_serializer.serialize_transaction(transaction)
        assert serialized.hex() == "085c120e00018ee90ff6181f3761632000001a208049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f82a200139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1388094ebdc0340f093094a0f746573742064617461206669656c64520d6c6f63616c2d746573746e657458026240c74c3e5b276c32ae72ab1a0bac17939d7577e55e1467cd3b4d4a45b04ad3957e94a855657b42601384819502882d559dd5f8e31a93097e113fe9fa82615151046802722032a3f14cf53c4d0543954f6cf1bda0369d13e661dec095107627dc0f6d33612f7a4000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"

        tx_hash = self.transaction_computer.compute_transaction_hash(transaction)
        assert tx_hash.hex() == "0896751a0c0eb3316041295ae1d71a4200aef5b359609b0cc181cae4e22b531e"

    def test_sign_transaction_by_hash(self):
        tx = Transaction(
            sender="moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8",
            receiver="moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk",
            value=0,
            gas_limit=50000,
            version=2,
            options=1,
            chain_id="integration tests chain ID",
            nonce=89
        )
        serialized = self.transaction_computer.compute_hash_for_signing(tx)
        tx.signature = self.alice.secret_key.sign(serialized)

        assert tx.signature.hex() == "e55e641a75b4357269bb576ec597aac8ef9650c8d37a75d4fc6c67ef78703e9dc8de17dca8a3b207610e202a3faea80f92f6e5b7b95d4fb91136d97c9167dc08"



from typing import List, Optional, Protocol

from drt_sdk_core.interfaces import (IAddress, IGasLimit, IGasPrice,
                                            INonce, ITransactionValue)
from drt_sdk_core.serializer import arg_to_string, args_to_strings
from drt_sdk_core.transaction_builders.transaction_builder import (
    ITransactionBuilderConfiguration, TransactionBuilder)


class IDCDTIssueConfiguration(ITransactionBuilderConfiguration, Protocol):
    gas_limit_dcdt_issue: IGasLimit
    issue_cost: ITransactionValue
    dcdt_contract_address: IAddress


class DCDTIssueBuilder(TransactionBuilder):
    def __init__(self,
                 config: IDCDTIssueConfiguration,
                 issuer: IAddress,
                 token_name: str,
                 token_ticker: str,
                 initial_supply: int,
                 num_decimals: int,
                 can_freeze: bool = False,
                 can_wipe: bool = False,
                 can_pause: bool = False,
                 can_mint: bool = False,
                 can_burn: bool = False,
                 can_change_owner: bool = False,
                 can_upgrade: bool = False,
                 can_add_special_roles: bool = False,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.gas_limit_dcdt_issue = config.gas_limit_dcdt_issue

        self.sender = issuer
        self.receiver = config.dcdt_contract_address
        self.value = config.issue_cost

        self.token_name = token_name
        self.token_ticker = token_ticker
        self.initial_supply = initial_supply
        self.num_decimals = num_decimals
        self.can_freeze = can_freeze
        self.can_wipe = can_wipe
        self.can_pause = can_pause
        self.can_mint = can_mint
        self.can_burn = can_burn
        self.can_change_owner = can_change_owner
        self.can_upgrade = can_upgrade
        self.can_add_special_roles = can_add_special_roles

    def _estimate_execution_gas(self) -> IGasLimit:
        return self.gas_limit_dcdt_issue

    def _build_payload_parts(self) -> List[str]:
        # the `canUpgrade` and `canAddSpecialRoles` attributes are set to `True` if not explicitly set to `False`
        return [
            "issue",
            arg_to_string(self.token_name),
            arg_to_string(self.token_ticker),
            arg_to_string(self.initial_supply),
            arg_to_string(self.num_decimals),
            *(args_to_strings(["canFreeze", "true"]) if self.can_freeze else []),
            *(args_to_strings(["canWipe", "true"]) if self.can_wipe else []),
            *(args_to_strings(["canPause", "true"]) if self.can_pause else []),
            *(args_to_strings(["canMint", "true"]) if self.can_mint else []),
            *(args_to_strings(["canBurn", "true"]) if self.can_burn else []),
            *(args_to_strings(["canChangeOwner", "true"]) if self.can_change_owner else []),
            *(args_to_strings(["canUpgrade", str(self.can_upgrade).lower()])),
            *(args_to_strings(["canAddSpecialRoles", str(self.can_add_special_roles).lower()]))
        ]



from typing import Protocol, Sequence

from drt_sdk_core.address import Address
from drt_sdk_core.constants import DELEGATION_MANAGER_SC_ADDRESS
from drt_sdk_core.errors import ErrListsLengthMismatch
from drt_sdk_core.interfaces import IAddress, IValidatorPublicKey
from drt_sdk_core.serializer import arg_to_string
from drt_sdk_core.transaction import Transaction
from drt_sdk_core.transaction_factories.transaction_builder import \
    TransactionBuilder


class IConfig(Protocol):
    chain_id: str
    min_gas_limit: int
    gas_limit_per_byte: int
    gas_limit_stake: int
    gas_limit_unstake: int
    gas_limit_unbond: int
    gas_limit_create_delegation_contract: int
    gas_limit_delegation_operations: int
    additional_gas_limit_per_validator_node: int
    additional_gas_for_delegation_operations: int


class DelegationTransactionsFactory:
    def __init__(self, config: IConfig) -> None:
        self.config = config

    def create_transaction_for_new_delegation_contract(self,
                                                       sender: IAddress,
                                                       total_delegation_cap: int,
                                                       service_fee: int,
                                                       amount: int) -> Transaction:
        parts = [
            "createNewDelegationContract",
            arg_to_string(total_delegation_cap),
            arg_to_string(service_fee)
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=Address.new_from_bech32(DELEGATION_MANAGER_SC_ADDRESS),
            data_parts=parts,
            gas_limit=self.config.gas_limit_create_delegation_contract + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True,
            amount=amount
        ).build()

        return transaction

    def create_transaction_for_adding_nodes(self,
                                            sender: IAddress,
                                            delegation_contract: IAddress,
                                            public_keys: Sequence[IValidatorPublicKey],
                                            signed_messages: Sequence[bytes]) -> Transaction:
        if len(public_keys) != len(signed_messages):
            raise ErrListsLengthMismatch("The number of public keys should match the number of signed messages")

        parts = ["addNodes"]
        for i in range(len(public_keys)):
            parts.append(public_keys[i].hex())
            parts.append(signed_messages[i].hex())

        num_nodes = len(public_keys)

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self._compute_execution_gas_limit_for_nodes_management(num_nodes),
            add_data_movement_gas=True
        ).build()

        return transaction

    def _compute_execution_gas_limit_for_nodes_management(self, num_nodes: int) -> int:
        return self.config.gas_limit_delegation_operations + num_nodes * self.config.additional_gas_limit_per_validator_node

    def create_transaction_for_removing_nodes(self,
                                              sender: IAddress,
                                              delegation_contract: IAddress,
                                              public_keys: Sequence[IValidatorPublicKey]) -> Transaction:
        num_nodes = len(public_keys)

        parts = ["removeNodes"]
        for public_key in public_keys:
            parts.append(public_key.hex())

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self._compute_execution_gas_limit_for_nodes_management(num_nodes),
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_staking_nodes(self,
                                             sender: IAddress,
                                             delegation_contract: IAddress,
                                             public_keys: Sequence[IValidatorPublicKey]) -> Transaction:
        num_nodes = len(public_keys)

        parts = ["stakeNodes"]
        for public_key in public_keys:
            parts.append(public_key.hex())

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.gas_limit_stake + num_nodes * self.config.additional_gas_limit_per_validator_node,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_unbonding_nodes(self,
                                               sender: IAddress,
                                               delegation_contract: IAddress,
                                               public_keys: Sequence[IValidatorPublicKey]) -> Transaction:
        num_nodes = len(public_keys)

        parts = ["unBondNodes"]
        for public_key in public_keys:
            parts.append(public_key.hex())

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.gas_limit_unbond + num_nodes * self.config.additional_gas_limit_per_validator_node,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_unstaking_nodes(self,
                                               sender: IAddress,
                                               delegation_contract: IAddress,
                                               public_keys: Sequence[IValidatorPublicKey]) -> Transaction:
        num_nodes = len(public_keys)

        parts = ["unStakeNodes"]
        for public_key in public_keys:
            parts.append(public_key.hex())

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.gas_limit_unstake + num_nodes * self.config.additional_gas_limit_per_validator_node,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_unjailing_nodes(self,
                                               sender: IAddress,
                                               delegation_contract: IAddress,
                                               public_keys: Sequence[IValidatorPublicKey]) -> Transaction:
        num_nodes = len(public_keys)

        parts = ["unJailNodes"]
        for public_key in public_keys:
            parts.append(public_key.hex())

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self._compute_execution_gas_limit_for_nodes_management(num_nodes),
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_changing_service_fee(self,
                                                    sender: IAddress,
                                                    delegation_contract: IAddress,
                                                    service_fee: int) -> Transaction:
        parts = [
            "changeServiceFee",
            arg_to_string(service_fee)
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_modifying_delegation_cap(self,
                                                        sender: IAddress,
                                                        delegation_contract: IAddress,
                                                        delegation_cap: int) -> Transaction:
        parts = [
            "modifyTotalDelegationCap",
            arg_to_string(delegation_cap)
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_setting_automatic_activation(self,
                                                            sender: IAddress,
                                                            delegation_contract: IAddress) -> Transaction:
        parts = [
            "setAutomaticActivation",
            arg_to_string('true')
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_unsetting_automatic_activation(self,
                                                              sender: IAddress,
                                                              delegation_contract: IAddress) -> Transaction:
        parts = [
            "setAutomaticActivation",
            arg_to_string('false')
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_setting_cap_check_on_redelegate_rewards(self,
                                                                       sender: IAddress,
                                                                       delegation_contract: IAddress) -> Transaction:
        parts = [
            "setCheckCapOnReDelegateRewards",
            arg_to_string('true')
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_unsetting_cap_check_on_redelegate_rewards(self,
                                                                         sender: IAddress,
                                                                         delegation_contract: IAddress) -> Transaction:
        parts = [
            "setCheckCapOnReDelegateRewards",
            arg_to_string('false')
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_setting_metadata(self,
                                                sender: IAddress,
                                                delegation_contract: IAddress,
                                                name: str,
                                                website: str,
                                                identifier: str) -> Transaction:
        parts = [
            "setMetaData",
            arg_to_string(name),
            arg_to_string(website),
            arg_to_string(identifier)
        ]

        transaction = TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=parts,
            gas_limit=self.config.gas_limit_delegation_operations + self.config.additional_gas_for_delegation_operations,
            add_data_movement_gas=True
        ).build()

        return transaction

    def create_transaction_for_delegating(self,
                                          sender: IAddress,
                                          delegation_contract: IAddress,
                                          amount: int) -> Transaction:
        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=["delegate"],
            gas_limit=12000000,
            add_data_movement_gas=False,
            amount=amount
        ).build()

    def create_transaction_for_claiming_rewards(self,
                                                sender: IAddress,
                                                delegation_contract: IAddress) -> Transaction:
        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=["claimRewards"],
            gas_limit=6000000,
            add_data_movement_gas=False,
            amount=0
        ).build()

    def create_transaction_for_redelegating_rewards(self,
                                                    sender: IAddress,
                                                    delegation_contract: IAddress) -> Transaction:
        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=["reDelegateRewards"],
            gas_limit=12000000,
            add_data_movement_gas=False,
            amount=0
        ).build()

    def create_transaction_for_undelegating(self,
                                            sender: IAddress,
                                            delegation_contract: IAddress,
                                            amount: int) -> Transaction:
        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=["unDelegate", arg_to_string(amount)],
            gas_limit=12000000,
            add_data_movement_gas=False,
            amount=0
        ).build()

    def create_transaction_for_withdrawing(self,
                                           sender: IAddress,
                                           delegation_contract: IAddress) -> Transaction:
        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=delegation_contract,
            data_parts=["withdraw"],
            gas_limit=12000000,
            add_data_movement_gas=False,
            amount=0
        ).build()



from drt_sdk_core.account import AccountNonceHolder
from drt_sdk_core.address import (Address, AddressComputer,
                                         AddressFactory)
from drt_sdk_core.code_metadata import CodeMetadata
from drt_sdk_core.contract_query import ContractQuery
from drt_sdk_core.contract_query_builder import ContractQueryBuilder
from drt_sdk_core.message import Message, MessageComputer
from drt_sdk_core.token_payment import TokenPayment
from drt_sdk_core.tokens import (Token, TokenComputer,
                                        TokenIdentifierParts, TokenTransfer)
from drt_sdk_core.transaction import Transaction, TransactionComputer
from drt_sdk_core.transaction_payload import TransactionPayload

__all__ = [
    "AccountNonceHolder", "Address", "AddressFactory", "AddressComputer",
    "Transaction", "TransactionPayload", "TransactionComputer",
    "Message", "MessageComputer", "CodeMetadata", "TokenPayment",
    "ContractQuery", "ContractQueryBuilder",
    "Token", "TokenComputer", "TokenTransfer", "TokenIdentifierParts"
]





from typing import List, Optional, Sequence

from drt_sdk_core.interfaces import IAddress


class ContractQuery:
    def __init__(self,
                 contract: IAddress,
                 function: str,
                 encoded_arguments: List[str] = [],
                 value: Optional[int] = None,
                 caller: Optional[IAddress] = None) -> None:
        self.contract: IAddress = contract
        self.function: str = function
        self.encoded_arguments = encoded_arguments
        self.value: int = value or 0
        self.caller: Optional[IAddress] = caller

    def get_contract(self) -> IAddress:
        return self.contract

    def get_function(self) -> str:
        return self.function

    def get_encoded_arguments(self) -> Sequence[str]:
        return self.encoded_arguments

    def get_caller(self) -> Optional[IAddress]:
        return self.caller

    def get_value(self) -> int:
        return self.value



from typing import Protocol

import drt_sdk_core.proto.transaction_pb2 as ProtoTransaction
from drt_sdk_core.address import Address
from drt_sdk_core.codec import encode_unsigned_number


class ITransaction(Protocol):
    sender: str
    receiver: str
    gas_limit: int
    chain_id: str
    gas_price: int
    sender_username: str
    receiver_username: str
    nonce: int
    value: int
    data: bytes
    version: int
    signature: bytes
    options: int
    guardian: str
    guardian_signature: bytes


class ProtoSerializer:
    def __init__(self) -> None:
        pass

    def serialize_transaction(self, transaction: ITransaction) -> bytes:
        receiver_pubkey = Address.new_from_bech32(transaction.receiver).get_public_key()
        sender_pubkey = Address.new_from_bech32(transaction.sender).get_public_key()

        proto_transaction = ProtoTransaction.Transaction()
        proto_transaction.Nonce = transaction.nonce
        proto_transaction.Value = self.serialize_transaction_value(transaction.value)
        proto_transaction.RcvAddr = receiver_pubkey
        proto_transaction.RcvUserName = transaction.receiver_username.encode()
        proto_transaction.SndAddr = sender_pubkey
        proto_transaction.SndUserName = transaction.sender_username.encode()
        proto_transaction.GasPrice = transaction.gas_price
        proto_transaction.GasLimit = transaction.gas_limit
        proto_transaction.Data = transaction.data
        proto_transaction.ChainID = transaction.chain_id.encode()
        proto_transaction.Version = transaction.version
        proto_transaction.Signature = transaction.signature
        proto_transaction.Options = transaction.options

        if transaction.guardian:
            guardian_address = transaction.guardian
            proto_transaction.GuardAddr = Address.new_from_bech32(guardian_address).get_public_key()
            proto_transaction.GuardSignature = transaction.guardian_signature

        encoded_tx: bytes = proto_transaction.SerializeToString()

        return encoded_tx

    def serialize_transaction_value(self, tx_value: int):
        if tx_value == 0:
            return bytes([0, 0])

        buffer = encode_unsigned_number(tx_value)
        buffer = bytes([0x00]) + buffer

        return buffer



from drt_sdk_core.message import Message, MessageComputer


def test_message_v1_serialize_for_signing():
    message = Message("test message".encode())
    message_computer = MessageComputer()
    serialized = message_computer.compute_bytes_for_signing(message)
    assert serialized.hex() == "2162d6271208429e6d3e664139e98ba7c5f1870906fb113e8903b1d3f531004d"



from drt_sdk_core.proto.transaction_serializer import ProtoSerializer
from drt_sdk_core.testutils.wallets import load_wallets
from drt_sdk_core.transaction import Transaction, TransactionComputer


class TestProtoSerializer:
    wallets = load_wallets()
    alice = wallets["alice"]
    bob = wallets["bob"]
    carol = wallets["carol"]
    proto_serializer = ProtoSerializer()
    transaction_computer = TransactionComputer()

    def test_serialize_tx_no_data_no_value(self):
        transaction = Transaction(
            sender=self.alice.label,
            receiver=self.bob.label,
            gas_limit=50000,
            chain_id="local-testnet",
            nonce=89,
            value=0,
        )
        transaction.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))

        serialized_transaction = self.proto_serializer.serialize_transaction(transaction)
        assert serialized_transaction.hex() == "0859120200001a208049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f82a200139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1388094ebdc0340d08603520d6c6f63616c2d746573746e657458026240de0b83647f9990b927ef9d3ff11ee0e6f2f50cc613c04c0e5b834f3e39eb0d2807cd664ce195713e1d9599e0a84cb56a400fe4e24e4e8d305e0d6dbc34597b0d"

    def test_serialize_tx_with_data_no_value(self):
        transaction = Transaction(
            sender=self.alice.label,
            receiver=self.bob.label,
            gas_limit=80000,
            chain_id="local-testnet",
            data=b"hello",
            nonce=90
        )
        transaction.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))

        serialized_transaction = self.proto_serializer.serialize_transaction(transaction)
        assert serialized_transaction.hex() == "085a120200001a208049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f82a200139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1388094ebdc034080f1044a0568656c6c6f520d6c6f63616c2d746573746e65745802624025fc0da0e2e1be76a217072ef27cee8fee2b8fdaa4154b128ec7565cf6bec61bf9c1df3a495fce4bafb06a5ed0f0d5b38dab3eaeb11ed862255938ae9d4dcc05"

    def test_serialize_tx_with_data_and_value(self):
        transaction = Transaction(
            sender=self.alice.label,
            receiver=self.bob.label,
            gas_limit=100000,
            chain_id="local-testnet",
            nonce=92,
            data=b"for the spaceship",
            value=123456789000000000000000000000
        )
        transaction.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))

        serialized_transaction = self.proto_serializer.serialize_transaction(transaction)
        assert serialized_transaction.hex() == "085c120e00018ee90ff6181f3761632000001a208049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f82a200139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1388094ebdc0340a08d064a11666f722074686520737061636573686970520d6c6f63616c2d746573746e6574580262401ff0a3e5da0b2e7d1160e8711456f5e1da865eb6607c0777511b621fac119e859ff900e7c56f0b0962c67edc75da10523ff5dc29a9b1ec8060685ad5dd1f3505"

    def test_serialize_tx_with_nonce_zero(self):
        transaction = Transaction(
            sender=self.alice.label,
            receiver=self.bob.label,
            chain_id="local-testnet",
            gas_limit=80000,
            nonce=0,
            value=0,
            data=b"hello",
            version=1
        )
        transaction.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))

        serialized_transaction = self.proto_serializer.serialize_transaction(transaction)
        assert serialized_transaction.hex() == "120200001a208049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f82a200139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1388094ebdc034080f1044a0568656c6c6f520d6c6f63616c2d746573746e657458016240d73e9c2f978d248eaba41c2453088f0e3488c02eb88ede7b1f40a22527a0f90c747056163a1625c127b3b6bd602a3a0cb607478a45d0a847e471911b3b87e805"

    def test_serialized_tx_with_usernames(self):
        transaction = Transaction(
            sender=self.carol.label,
            receiver=self.alice.label,
            gas_limit=50000,
            chain_id="T",
            nonce=204,
            value=1000000000000000000,
            sender_username="carol",
            receiver_username="alice"
        )
        transaction.signature = self.carol.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(transaction))

        serialized_transaction = self.proto_serializer.serialize_transaction(transaction)
        assert serialized_transaction.hex() == "08cc011209000de0b6b3a76400001a200139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e12205616c6963652a20b2a11555ce521e4944e09ab17549d85b487dcd26c84b5017a39e31a3670889ba32056361726f6c388094ebdc0340d08603520154580262403f3f645f9ac38142c1087a76b855e48ed0aa44185146df5f3317b8887e7b59f195d6093d3c2b30fb693a27e3175a199ef04a6e650101ab3f4b3649f0fe2f5204"




from drt_sdk_core.account import AccountNonceHolder


def test_account_nonce_holder():
    account = AccountNonceHolder(41)
    account.nonce = 42
    assert account.get_nonce_then_increment() == 42
    assert account.get_nonce_then_increment() == 43

    account.increment_nonce()
    account.increment_nonce()
    account.increment_nonce()
    assert account.nonce == 47



import base64
import json
from typing import Any, Dict, Optional

from drt_sdk_core.address import Address
from drt_sdk_core.errors import ErrInvalidRelayerV1BuilderArguments
from drt_sdk_core.interfaces import (IAddress, INetworkConfig, INonce,
                                            ITransactionOptions,
                                            ITransactionVersion)
from drt_sdk_core.transaction import Transaction


class RelayedTransactionV1Builder:
    def __init__(
        self,
        inner_transaction: Optional[Transaction] = None,
        relayer_address: Optional[IAddress] = None,
        relayer_nonce: Optional[INonce] = None,
        network_config: Optional[INetworkConfig] = None,
        relayed_transaction_options: Optional[ITransactionOptions] = None,
        relayed_transaction_version: Optional[ITransactionVersion] = None,
        relayed_transaction_guardian: Optional[IAddress] = None,
    ) -> None:
        self.inner_transaction = inner_transaction
        self.relayer_address = relayer_address
        self.relayer_nonce = relayer_nonce
        self.network_config = network_config
        self.relayed_transaction_options = relayed_transaction_options
        self.relayed_transaction_version = relayed_transaction_version
        self.relayed_transaction_guardian = relayed_transaction_guardian

    def set_inner_transaction(self, transaction: Transaction) -> None:
        self.inner_transaction = transaction

    def set_network_config(self, network_config: INetworkConfig) -> None:
        self.network_config = network_config

    def set_relayer_address(self, relayer_address: IAddress) -> None:
        self.relayer_address = relayer_address

    def set_relayer_nonce(self, relayer_nonce: INonce) -> None:
        self.relayer_nonce = relayer_nonce

    def set_relayed_transaction_version(
        self, relayed_transaction_version: ITransactionVersion
    ) -> None:
        self.relayed_transaction_version = relayed_transaction_version

    def set_relayed_transaction_options(
        self, relayed_transaction_options: ITransactionOptions
    ) -> None:
        self.relayed_transaction_options = relayed_transaction_options

    def set_relayed_transaction_guardian(
        self, relayed_transaction_guardian: IAddress
    ) -> None:
        self.relayed_transaction_guardian = relayed_transaction_guardian

    def build(self) -> Transaction:
        if (
            not self.inner_transaction
            or not self.network_config
            or not self.relayer_address
            or not self.inner_transaction.signature
        ):
            raise ErrInvalidRelayerV1BuilderArguments()

        serialized_transaction = self._prepare_inner_transaction()
        data = f"relayedTx@{serialized_transaction.encode().hex()}"

        gas_limit = (
            self.network_config.min_gas_limit
            + self.network_config.gas_per_data_byte * len(data)
            + self.inner_transaction.gas_limit
        )
        relayed_transaction = Transaction(
            chain_id=self.network_config.chain_id,
            sender=self.relayer_address.to_bech32(),
            receiver=self.inner_transaction.sender,
            value=0,
            nonce=self.relayer_nonce,
            gas_limit=gas_limit,
            data=data.encode(),
            version=self.relayed_transaction_version,
            options=self.relayed_transaction_options,
            guardian=self.relayed_transaction_guardian.to_bech32() if self.relayed_transaction_guardian else "",
        )

        if self.relayer_nonce:
            self.set_relayer_nonce(self.relayer_nonce)

        return relayed_transaction

    def _prepare_inner_transaction(self) -> str:
        if not self.inner_transaction:
            return ""

        sender = Address.new_from_bech32(self.inner_transaction.sender).to_hex()
        receiver = Address.new_from_bech32(self.inner_transaction.receiver).to_hex()

        tx: Dict[str, Any] = {
            "nonce": self.inner_transaction.nonce,
            "sender": base64.b64encode(bytes.fromhex(sender)).decode(),
            "receiver": base64.b64encode(bytes.fromhex(receiver)).decode(),
            "value": int(self.inner_transaction.value.__str__(), 10),
            "gasPrice": self.inner_transaction.gas_price,
            "gasLimit": self.inner_transaction.gas_limit,
            "data": base64.b64encode(self.inner_transaction.data).decode(),
            "signature": base64.b64encode(self.inner_transaction.signature).decode(),
            "chainID": base64.b64encode(
                self.inner_transaction.chain_id.encode()
            ).decode(),
            "version": self.inner_transaction.version,
        }

        if self.inner_transaction.options:
            tx["options"] = self.inner_transaction.options

        if self.inner_transaction.guardian:
            guardian = Address.new_from_bech32(self.inner_transaction.guardian).to_hex()
            tx["guardian"] = base64.b64encode(bytes.fromhex(guardian)).decode()

        if self.inner_transaction.guardian_signature:
            tx["guardianSignature"] = base64.b64encode(self.inner_transaction.guardian_signature).decode()

        if self.inner_transaction.sender_username:
            tx["sndUserName"] = base64.b64encode(self.inner_transaction.sender_username.encode()).decode()

        if self.inner_transaction.receiver_username:
            tx[f"rcvUserName"] = base64.b64encode(self.inner_transaction.receiver_username.encode()).decode()

        return json.dumps(tx, separators=(",", ":"))




from typing import Protocol


class IAddress(Protocol):
    def to_bech32(self) -> str: ...

    def to_hex(self) -> str: ...


class ITransaction(Protocol):
    sender: str
    receiver: str
    gas_limit: int
    chain_id: str
    nonce: int
    value: int
    sender_username: str
    receiver_username: str
    gas_price: int
    data: bytes
    version: int
    options: int
    guardian: str
    signature: bytes
    guardian_signature: bytes


class IMessage(Protocol):
    data: bytes
    signature: bytes


class IToken(Protocol):
    identifier: str
    nonce: int


class ITokenTransfer(Protocol):
    @property
    def token(self) -> IToken:
        ...

    amount: int


class ITokenIdentifierParts(Protocol):
    ticker: str
    random_sequence: str
    nonce: int


INonce = int
IGasPrice = int
IGasLimit = int
IChainID = str
ITransactionVersion = int
ITransactionOptions = int
ISignature = bytes
ITokenIdentifier = str
IGasPerDataByte = int


class ITokenPayment(Protocol):
    token_identifier: ITokenIdentifier
    token_nonce: INonce
    amount_as_integer: int

    def is_rewa(self) -> bool: ...
    def is_fungible(self) -> bool: ...


class ITransactionValue(Protocol):
    def __str__(self) -> str: ...


class ITransactionPayload(Protocol):
    data: bytes
    def encoded(self) -> str: ...
    def length(self) -> int: ...


class ICodeMetadata(Protocol):
    def serialize(self) -> bytes: ...


class INetworkConfig(Protocol):
    min_gas_limit: IGasLimit
    gas_per_data_byte: IGasPerDataByte
    gas_price_modifier: float
    chain_id: IChainID


class IValidatorPublicKey(Protocol):
    def hex(self) -> str:
        ...



from typing import Protocol, Sequence

from drt_sdk_core.interfaces import IAddress


class ITransactionOnNetwork(Protocol):
    @property
    def contract_results(self) -> 'IContractResults': ...
    @property
    def logs(self) -> 'ITransactionLogs': ...


class IContractResults(Protocol):
    @property
    def items(self) -> Sequence['IContractResultItem']: ...


class IContractResultItem(Protocol):
    @property
    def logs(self) -> 'ITransactionLogs': ...


class ITransactionLogs(Protocol):
    @property
    def events(self) -> Sequence['ITransactionEvent']: ...


class ITransactionEvent(Protocol):
    @property
    def address(self) -> IAddress: ...

    @property
    def identifier(self) -> str: ...

    @property
    def topics(self) -> Sequence['ITransactionEventTopic']: ...

    @property
    def data(self) -> str: ...


class ITransactionEventTopic(Protocol):
    @property
    def raw(self) -> bytes: ...



from typing import Any, List, Optional

from drt_sdk_core.address import Address
from drt_sdk_core.errors import (ErrInvalidGasLimitForInnerTransaction,
                                        ErrInvalidRelayerV2BuilderArguments)
from drt_sdk_core.interfaces import (IAddress, IGasLimit,
                                            INetworkConfig, INonce)
from drt_sdk_core.serializer import args_to_string
from drt_sdk_core.transaction import Transaction
from drt_sdk_core.transaction_payload import TransactionPayload


class RelayedTransactionV2Builder:
    def __init__(
            self, inner_transaction: Optional[Transaction] = None,
            inner_transaction_gas_limit: Optional[IGasLimit] = None,
            relayer_address: Optional[IAddress] = None,
            relayer_nonce: Optional[INonce] = None,
            network_config: Optional[INetworkConfig] = None) -> None:
        self.inner_transaction = inner_transaction
        self.inner_transaction_gas_limit = inner_transaction_gas_limit
        self.relayer_address = relayer_address
        self.relayer_nonce = relayer_nonce
        self.network_config = network_config

    def set_inner_transaction(self, transaction: Transaction) -> None:
        self.inner_transaction = transaction

    def set_inner_transaction_gas_limit(self, gas_limit: IGasLimit) -> None:
        self.inner_transaction_gas_limit = gas_limit

    def set_network_config(self, network_config: INetworkConfig) -> None:
        self.network_config = network_config

    def set_relayer_address(self, relayer_address: IAddress) -> None:
        self.relayer_address = relayer_address

    def set_relayer_nonce(self, relayer_nonce: INonce) -> None:
        self.relayer_nonce = relayer_nonce

    def build(self) -> Transaction:
        if (
            not self.inner_transaction
            or not self.network_config
            or not self.relayer_address
            or not self.inner_transaction.signature
            or not self.inner_transaction_gas_limit
        ):
            raise ErrInvalidRelayerV2BuilderArguments()

        if self.inner_transaction.gas_limit:
            raise ErrInvalidGasLimitForInnerTransaction()

        arguments: List[Any] = [
            Address.new_from_bech32(self.inner_transaction.receiver),
            self.inner_transaction.nonce,
            self.inner_transaction.data,
            self.inner_transaction.signature
        ]

        data = f"relayedTxV2@{args_to_string(arguments)}"
        payload = TransactionPayload.from_str(data)

        relayed_transaction = Transaction(
            sender=self.relayer_address.to_bech32(),
            receiver=self.inner_transaction.sender,
            value=0,
            gas_limit=self.inner_transaction_gas_limit + self.network_config.min_gas_limit + self.network_config.gas_per_data_byte * payload.length(),
            chain_id=self.network_config.chain_id,
            data=payload.data,
            version=self.inner_transaction.version,
            options=self.inner_transaction.options
        )

        if self.relayer_nonce:
            relayed_transaction.nonce = self.relayer_nonce

        return relayed_transaction



from drt_sdk_core import Address
from drt_sdk_core.codec import encode_unsigned_number
from drt_sdk_core.transaction_factories.transactions_factory_config import \
    TransactionsFactoryConfig
from drt_sdk_core.transaction_parsers.token_operations_outcome_parser import \
    TokenOperationsOutcomeParser
from drt_sdk_core.transaction_parsers.transaction_on_network_wrapper import (
    ContractResultsWrapper, TransactionEventTopicWrapper,
    TransactionEventWrapper, TransactionLogsWrapper,
    TransactionOnNetworkWrapper)

frank = Address.new_from_bech32("moa1kdl46yctawygtwg2k462307dmz2v55c605737dp3zkxh04sct7asdueccu")
grace = Address.new_from_bech32("moa1r69gk66fmedhhcg24g2c5kn2f2a5k4kvpr6jfw67dn2lyydd8cfsruuk3f")
parser = TokenOperationsOutcomeParser(TransactionsFactoryConfig("D"))


def test_parse_issue_fungible():
    transaction = TransactionOnNetworkWrapper(
        ContractResultsWrapper([]),
        TransactionLogsWrapper([
            TransactionEventWrapper(
                address=frank,
                identifier="issue",
                topics=[TransactionEventTopicWrapper("FOOBAR".encode())],
                data=""
            )
        ])
    )

    outcome = parser.parse_issue_fungible(transaction)
    assert outcome.token_identifier == "FOOBAR"


def test_parse_set_special_role():
    transaction = TransactionOnNetworkWrapper(
        ContractResultsWrapper([]),
        TransactionLogsWrapper([
            TransactionEventWrapper(
                address=grace,
                identifier="DCDTSetRole",
                topics=[
                    TransactionEventTopicWrapper("FOOBAR".encode()),
                    TransactionEventTopicWrapper("".encode()),
                    TransactionEventTopicWrapper("".encode()),
                    TransactionEventTopicWrapper("DCDTRoleLocalMint".encode()),
                    TransactionEventTopicWrapper("DCDTRoleLocalBurn".encode())
                ],
                data=""
            )
        ])
    )

    outcome = parser.parse_set_special_role(transaction)
    assert outcome.token_identifier == "FOOBAR"
    assert outcome.roles == ["DCDTRoleLocalMint", "DCDTRoleLocalBurn"]
    assert outcome.user_address == grace.to_bech32()


def test_parse_local_mint():
    transaction = TransactionOnNetworkWrapper(
        ContractResultsWrapper([]),
        TransactionLogsWrapper([
            TransactionEventWrapper(
                address=grace,
                identifier="DCDTLocalMint",
                topics=[
                    TransactionEventTopicWrapper("FOOBAR".encode()),
                    TransactionEventTopicWrapper("".encode()),
                    TransactionEventTopicWrapper(encode_unsigned_number(200))
                ],
                data=""
            )
        ])
    )

    outcome = parser.parse_local_mint(transaction)
    assert outcome.token_identifier == "FOOBAR"
    assert outcome.nonce == 0
    assert outcome.minted_supply == 200
    assert outcome.user_address == grace.to_bech32()


def test_parse_nft_create():
    transaction = TransactionOnNetworkWrapper(
        ContractResultsWrapper([]),
        TransactionLogsWrapper([
            TransactionEventWrapper(
                address=grace,
                identifier="DCDTNFTCreate",
                topics=[
                    TransactionEventTopicWrapper("FOOBAR".encode()),
                    TransactionEventTopicWrapper(encode_unsigned_number(42)),
                    TransactionEventTopicWrapper(encode_unsigned_number(1))
                ],
                data=""
            )
        ])
    )

    outcome = parser.parse_nft_create(transaction)
    assert outcome.token_identifier == "FOOBAR"
    assert outcome.nonce == 42
    assert outcome.initial_quantity == 1



from typing import List, Protocol, Sequence

from drt_sdk_core.interfaces import IAddress, ITokenTransfer
from drt_sdk_core.serializer import arg_to_string, args_to_strings


class ITokenComputer(Protocol):
    def extract_identifier_from_extended_identifier(self, identifier: str) -> str:
        ...


class TokenTransfersDataBuilder:
    def __init__(self, token_computer: ITokenComputer) -> None:
        self.token_computer = token_computer

    def build_args_for_dcdt_transfer(self, transfer: ITokenTransfer) -> List[str]:
        args = ["DCDTTransfer"]
        args.extend(args_to_strings([transfer.token.identifier, transfer.amount]))

        return args

    def build_args_for_single_dcdt_nft_transfer(self, transfer: ITokenTransfer, receiver: IAddress) -> List[str]:
        args = ["DCDTNFTTransfer"]
        token = transfer.token
        identifier = self.token_computer.extract_identifier_from_extended_identifier(token.identifier)
        args.extend(args_to_strings([identifier, token.nonce, transfer.amount]))
        args.append(receiver.to_hex())

        return args

    def build_args_for_multi_dcdt_nft_transfer(self, receiver: IAddress, transfers: Sequence[ITokenTransfer]) -> List[str]:
        args = ["MultiDCDTNFTTransfer", receiver.to_hex(), arg_to_string(len(transfers))]

        for transfer in transfers:
            identifier = self.token_computer.extract_identifier_from_extended_identifier(transfer.token.identifier)
            args.extend(args_to_strings([identifier, transfer.token.nonce, transfer.amount]))

        return args



from drt_sdk_core.codec import encode_signed_number

test_vectors_1 = [
    [-1, 0xFF],
    [1, 0x01],
    [2, 0x02],
    [-2, 0xFE],
    [3, 0x03],
    [-3, 0xFD],
    [4, 0x04],
    [-4, 0xFC],
]

test_vectors_2 = [
    [-1, 0xFF],
    [1, 0x01],
    [2, 0x02],
    [-2, 0xFE],
    [3, 0x03],
    [-3, 0xFD],
    [4, 0x04],
    [-4, 0xFC],
    [5, 0x05],
    [-5, 0xFB],
    [6, 0x06],
    [-6, 0xFA],
]

test_vectors_3 = [
    [125, 0x7D],
    [-125, 0x83],
    [126, 0x7E],
    [-126, 0x82],
    [127, 0x7F],
    [-127, 0x81],
    [-128, 0x80],
]

test_vectors_4 = [
    [128, [0x00, 0x80]],
    [129, [0x00, 0x81]],
    [-129, [0xFF, 0x7F]],
    [130, [0x00, 0x82]],
    [-130, [0xFF, 0x7E]],
    [253, [0x00, 0xFD]],
    [256, [0x01, 0x00]],
    [-256, [0xFF, 0x00]],
    [-257, [0xFE, 0xFF]],
    [258, [0x01, 0x02]],
]


def test_encode_signed_number():
    assert encode_signed_number(-256) == bytes([0xFF, 0x00])
    assert encode_signed_number(-0x11) == bytes([0xEF])
    assert encode_signed_number(-128) == bytes([0x80])
    assert encode_signed_number(-1) == bytes([0xFF])
    assert encode_signed_number(0) == bytes([])
    assert encode_signed_number(1) == bytes([0x01])
    assert encode_signed_number(256) == bytes([0x01, 0x00])
    assert encode_signed_number(127) == bytes([0x7F])
    assert encode_signed_number(0x11) == bytes([0x11])
    assert encode_signed_number(255) == bytes([0x00, 0xFF])

    for input_data, expected_data in test_vectors_1:
        assert encode_signed_number(input_data) == bytes([expected_data])

    for input_data, expected_data in test_vectors_2:
        assert encode_signed_number(input_data) == bytes([expected_data])

    for input_data, expected_data in test_vectors_3:
        assert encode_signed_number(input_data) == bytes([expected_data])

    for input_data, expected_data in test_vectors_4:
        assert encode_signed_number(input_data) == bytes(expected_data)  # type: ignore



import logging
from typing import Protocol, Tuple

from Cryptodome.Hash import keccak

from drt_sdk_core import bech32
from drt_sdk_core.constants import DEFAULT_HRP, METACHAIN_ID
from drt_sdk_core.errors import ErrBadAddress, ErrBadPubkeyLength

SC_HEX_PUBKEY_PREFIX = "0" * 16
PUBKEY_LENGTH = 32
PUBKEY_STRING_LENGTH = PUBKEY_LENGTH * 2  # hex-encoded
BECH32_LENGTH = 62

logger = logging.getLogger("address")


class IAddress(Protocol):
    def get_public_key(self) -> bytes:
        ...

    def get_hrp(self) -> str:
        ...


class Address:
    def __init__(self, pubkey: bytes, hrp: str) -> None:
        if len(pubkey) != PUBKEY_LENGTH:
            raise ErrBadPubkeyLength(len(pubkey), PUBKEY_LENGTH)

        self.pubkey = bytes(pubkey)
        self.hrp = hrp

    @classmethod
    def new_from_bech32(cls, value: str) -> 'Address':
        hrp, pubkey = _decode_bech32(value)
        return cls(pubkey, hrp)

    @classmethod
    def from_bech32(cls, value: str) -> 'Address':
        """The `from_bech32()` method is deprecated. Please use `new_from_bech32()` instead"""
        return Address.new_from_bech32(value)

    @classmethod
    def new_from_hex(cls, value: str, hrp: str) -> 'Address':
        pubkey = bytes.fromhex(value)
        return cls(pubkey, hrp)

    @classmethod
    def from_hex(cls, value: str, hrp: str) -> 'Address':
        """The `from_hex()` method is deprecated. Please use `new_from_hex()` instead"""
        return Address.new_from_hex(value, hrp)

    def to_hex(self) -> str:
        return self.pubkey.hex()

    def hex(self) -> str:
        """The `hex()` method is deprecated. Please use `to_hex()` instead"""
        return self.to_hex()

    def to_bech32(self) -> str:
        converted = bech32.convertbits(self.pubkey, 8, 5)
        assert converted is not None
        encoded = bech32.bech32_encode(self.hrp, converted)
        return encoded

    def bech32(self) -> str:
        """The `bech32()` method is deprecated. Please us `to_bech32()` instead"""
        return self.to_bech32()

    def get_public_key(self) -> bytes:
        return self.pubkey

    def get_hrp(self) -> str:
        return self.hrp

    def is_smart_contract(self):
        return self.to_hex().startswith(SC_HEX_PUBKEY_PREFIX)

    # this will be removed in v1.0.0; it's here for compatibility reasons with the deprecated transaction builders
    # the transaction builders will also be removed in v1.0.0
    def serialize(self) -> bytes:
        logger.warning("The `serialize()` method is deprecated and will soon be removed")
        return self.get_public_key()


class AddressFactory:
    def __init__(self, hrp: str = DEFAULT_HRP) -> None:
        self.hrp = hrp

    def create_from_bech32(self, value: str) -> Address:
        hrp, pubkey = _decode_bech32(value)
        if hrp != self.hrp:
            raise ErrBadAddress(value)

        return Address(pubkey, hrp)

    def create_from_public_key(self, pubkey: bytes) -> Address:
        return Address(pubkey, self.hrp)

    def create_from_hex(self, value: str) -> Address:
        return Address.new_from_hex(value, self.hrp)


class AddressComputer:
    def __init__(self, number_of_shards: int = 3) -> None:
        self.number_of_shards = number_of_shards

    def compute_contract_address(self, deployer: IAddress, deployment_nonce: int) -> Address:
        """
        8 bytes of zero + 2 bytes for VM type + 20 bytes of hash(owner) + 2 bytes of shard(owner)
        """
        deployer_pubkey = deployer.get_public_key()
        nonce_bytes = deployment_nonce.to_bytes(8, byteorder="little")
        bytes_to_hash = deployer_pubkey + nonce_bytes
        contract_pubkey = keccak.new(digest_bits=256).update(bytes_to_hash).digest()
        contract_pubkey = bytes([0] * 8) + bytes([5, 0]) + contract_pubkey[10:30] + deployer_pubkey[30:]
        return Address(contract_pubkey, deployer.get_hrp())

    def get_shard_of_address(self, address: IAddress) -> int:
        return get_shard_of_pubkey(address.get_public_key(), self.number_of_shards)


def is_valid_bech32(value: str, expected_hrp: str) -> bool:
    hrp, value_bytes = bech32.bech32_decode(value)
    return hrp == expected_hrp and value_bytes is not None


def _decode_bech32(value: str) -> Tuple[str, bytes]:
    hrp, value_bytes = bech32.bech32_decode(value)
    if hrp is None or value_bytes is None:
        raise ErrBadAddress(value)

    decoded_bytes = bech32.convertbits(value_bytes, 5, 8, False)
    if decoded_bytes is None:
        raise ErrBadAddress(value)

    return hrp, bytes(bytearray(decoded_bytes))


def get_shard_of_pubkey(pubkey: bytes, number_of_shards: int) -> int:
    mask_high = int("11", 2)
    mask_low = int("01", 2)

    last_byte_of_pubkey = pubkey[31]

    if _is_pubkey_of_metachain(pubkey):
        return METACHAIN_ID

    shard = last_byte_of_pubkey & mask_high
    if shard > number_of_shards - 1:
        shard = last_byte_of_pubkey & mask_low

    return shard


def _is_pubkey_of_metachain(pubkey: bytes) -> bool:
    metachain_prefix = bytearray([0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    pubkey_prefix = pubkey[0:len(metachain_prefix)]
    if pubkey_prefix == bytes(metachain_prefix):
        return True

    zero_address = bytearray(32)
    if pubkey == bytes(zero_address):
        return True

    return False



from drt_sdk_core.address import DEFAULT_HRP, Address
from drt_sdk_core.interfaces import IAddress


class TransactionsFactoryConfig:
    def __init__(self, chain_id: str) -> None:
        # General-purpose configuration
        self.chain_id: str = chain_id
        self.address_hrp: str = DEFAULT_HRP
        self.min_gas_limit: int = 50_000
        self.gas_limit_per_byte: int = 1_500

        # Configuration for token operations
        self.gas_limit_issue: int = 60_000_000
        self.gas_limit_toggle_burn_role_globally: int = 60_000_000
        self.gas_limit_dcdt_local_mint: int = 300_000
        self.gas_limit_dcdt_local_burn: int = 300_000
        self.gas_limit_set_special_role: int = 60_000_000
        self.gas_limit_pausing: int = 60_000_000
        self.gas_limit_freezing: int = 60_000_000
        self.gas_limit_wiping: int = 60_000_000
        self.gas_limit_dcdt_nft_create: int = 3_000_000
        self.gas_limit_dcdt_nft_update_attributes: int = 1_000_000
        self.gas_limit_dcdt_nft_add_quantity: int = 1_000_000
        self.gas_limit_dcdt_nft_burn: int = 1_000_000
        self.gas_limit_store_per_byte: int = 50_000
        self.issue_cost: int = 50_000_000_000_000_000
        self.dcdt_contract_address: IAddress = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv")

        # Configuration for delegation operations
        self.gas_limit_stake = 5_000_000
        self.gas_limit_unstake = 5_000_000
        self.gas_limit_unbond = 5_000_000
        self.gas_limit_create_delegation_contract = 50_000_000
        self.gas_limit_delegation_operations = 1_000_000
        self.additional_gas_limit_per_validator_node = 6_000_000
        self.additional_gas_for_delegation_operations = 10_000_000

        # Configuration for token transfers
        self.gas_limit_dcdt_transfer = 200_000
        self.gas_limit_dcdt_nft_transfer = 200_000
        self.gas_limit_multi_dcdt_nft_transfer = 200_000
